#define G_LOG_DOMAIN "boilerplate-application"

#include "config.h"

#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include "boilerplate-application.h"
#include "boilerplate-window.h"
#include "boilerplate-preferences.h"

typedef struct
{
    BoilerplateWindow *window;

    GList *windows;

    GSettings *settings;

    gboolean csd;
    gboolean dark_mode;
} BoilerplateApplicationPrivate;


G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateApplication, boilerplate_application, GTK_TYPE_APPLICATION)

static const GOptionEntry option_entries[] = {
    { "version", '\0', 0, G_OPTION_ARG_NONE, NULL, "Print release version and exit", NULL },
    { "dark-mode", '\0', 0, G_OPTION_ARG_NONE, NULL, "Enable dark theme", NULL },
    //TODO: --traditional-mode ?
    { "disable-csd", 'c', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, NULL, "Disable Client Side Decorations (CSD)", NULL },
    { NULL }
};

/**
 * boilerplate_application_set_accelerator:
 * @application: The application (self)
 * @action_name: The name of the action, eg. app.preferences, win.show-help-overlay
 * @accel: The accelerator, eg. <Primary>N
 *
 * A helper function to quickly create an accelerator
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
void
boilerplate_application_set_accelerator (GtkApplication *application,
                                         const gchar    *action_name,
                                         const gchar    *accel)
{
    const gchar *vaccels[] = {
        accel,
        NULL
    };

    gtk_application_set_accels_for_action (application, action_name, vaccels);
}

/**
 * boilerplate_application_set_accelerators:
 * @application: The application (self)
 * @action_name: The name of the action, eg. app.preferences, win.show-help-overlay
 * @accels: An array of accelerator to ad for the @action_name
 *
 * A helper function to quickly create accelerators
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
void
boilerplate_application_set_accelerators (GtkApplication *application,
                                          const gchar    *action_name,
                                          const gchar   **accels)
{
    gtk_application_set_accels_for_action (application, action_name, accels);
}

/**
 * boilerplate_application_show_uri:
 * @user_data: The application
 * @uri: A string that contains the url to open
 *
 * A helper function for gtk_show_uri (), this function takes an application and
 * the uri to show and loads the proper screen, and checks to see if errors
 * occure. If so they are shown to the user a a dialog instead of in the
 * terminal.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_show_uri (GApplication *user_data,
                                  const gchar  *uri)
{
    g_debug ("Opening URI window for: %s", uri);

    GtkApplication *application;
    GtkWindow *window;
    GtkWidget *dialog;
    GError *error = NULL;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

    gtk_show_uri (window ? gtk_window_get_screen (GTK_WINDOW (window)) : gdk_screen_get_default (),
                  uri,
                  gtk_get_current_event_time (),
                  &error);

    if (error)
    {
        dialog = gtk_message_dialog_new (window ? GTK_WINDOW (window) : NULL,
                                         GTK_DIALOG_MODAL,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_OK,
                                         "There was an error displaying help: \n%s",
                                         error->message);

        g_signal_connect (G_OBJECT (dialog), "response",
                          G_CALLBACK (gtk_widget_destroy),
                          NULL);

        gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
        gtk_widget_show (dialog);
        g_error_free (error);
    }
}

/**
 * preferences_cb:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Opens the application's preferences.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
preferences_cb (GSimpleAction *action,
                GVariant      *parameter,
                gpointer       user_data)
{
    g_debug ("Launching preferences");

    GtkApplication *application;
    BoilerplateWindow *window;
    BoilerplatePreferences *preferences;

    application = GTK_APPLICATION (user_data);
    window = BOILERPLATE_WINDOW (gtk_application_get_active_window (application));

    preferences = boilerplate_preferences_new (window);

    gtk_window_present (GTK_WINDOW (preferences));
}

/**
 * help_cb:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Opens the application's help window.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
help_cb (GSimpleAction *action,
         GVariant      *parameter,
         gpointer       user_data)
{
    g_debug ("Launching help");

    boilerplate_application_show_uri (user_data,
                                      "help:" PACKAGE);
}

/**
 * about_cb:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Opens the application's about dialog.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
about_cb (GSimpleAction *action,
          GVariant      *parameter,
          gpointer       user_data)
{
    g_debug ("Launching about");

    GtkApplication *application;
    GtkWindow *window;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

	static const char *authors[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const char *artists[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const gchar *translators =
        "Julian Richen <julian@richen.io>";

	gtk_show_about_dialog (window,
		                   "program-name", g_get_application_name (),
		                   "version", PACKAGE_VERSION,
		                   "copyright", PACKAGE_COPYRIGHT,
		                   "comments", PACKAGE_SUMMARY,
		                   "authors", authors,
		                   "artists", artists,
		                   "translator-credits", translators,
		                   "website-label", "Website",
		                   "website", PACKAGE_URL_HOMEPAGE,
		                   "logo-icon-name", gtk_window_get_default_icon_name (),
		                   "wrap-license", TRUE,
		                   "license-type", GTK_LICENSE_GPL_3_0,
		                   NULL);
}

/**
 * quit_cb:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Action to quit the application.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
quit_cb (GSimpleAction *action,
         GVariant      *parameter,
         gpointer       user_data)
{
    g_debug ("Quiting application");

    g_application_quit (G_APPLICATION (user_data));
}

/**
 * action_entries[]
 *
 * Connects the the app-menu and menubar links in the respective menu to an
 * action. The "Keyboard Shortcuts" window is known as the help-overlay internally.
 * It is not listed here because the GtkShortcutsWindow with an id of help_overlay
 * is included as a resource in gresource via gtk/help-overlay.ui. Gtk will
 * automaticlly include and connect it with an action of win.show-help-overlay
 * and default shortcuts (Control-F1 and Control-?).
 */
static GActionEntry action_entries[] = {
    { "preferences", preferences_cb, NULL, NULL, NULL },
    { "help", help_cb, NULL, NULL, NULL },
    { "about", about_cb, NULL, NULL, NULL },
	{ "quit", quit_cb,  NULL, NULL, NULL },
};

/**
 * boilerplate_application_shutdown:
 * @application: The application
 *
 * Application Sequence: Final
 * Shutdown the application, all clean-up should be done here. I.e saving data,
 * states, etc...
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_shutdown (GApplication *application)
{
    g_debug ("Application shutting down");

    G_APPLICATION_CLASS (boilerplate_application_parent_class)->shutdown (application);
}

/**
 * boilerplate_application_activate:
 * @application: The application
 *
 * Application Sequence: 04
 * Activates the GtkApplicationWindow and presents the window to the user.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_activate (GApplication *application)
{
    g_debug ("Application activated");

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);

    /*
     * Create GtkApplicationWindow if needed
     * If you need multiple windows open (like you have multiple instances) you
     * should convert priv->window to a GList *windows and loop the list to
     * select the window
     */
    if (priv->window == NULL) {
	    priv->window = boilerplate_window_new (BOILERPLATE_APPLICATION (application));

        g_debug ("GtkApplicaitonWindow created and connected");
    }

    //TODO: Replace with plugin
    boilerplate_window_set_csd (BOILERPLATE_WINDOW (priv->window),
                                priv->csd);

    g_debug ("Presenting GtkApplicaitonWindow");

	gtk_window_present (GTK_WINDOW (priv->window));
}

/**
 * boilerplate_application_startup:
 * @application: The application
 *
 * Application Sequence: 03
 * Starts many of the applications features. It also sets information used
 * throughout the application like the application:
 * - name
 * - icon
 * - appmenu
 * - menubar
 * - action entries
 * - accelerators
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_startup (GApplication *application)
{
    g_debug ("Application startup");

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);

    /* Starting-up application */
	G_APPLICATION_CLASS (boilerplate_application_parent_class)->startup (application);

    /* Init menus */
    g_debug ("Creating application action entries");

    /* Add action callbacks */
	g_action_map_add_action_entries (G_ACTION_MAP (application),
                                     action_entries, G_N_ELEMENTS (action_entries),
                                     application);

    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.preferences", "<Primary>comma");
    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.help", "F1");
    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.about", "F2");
    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "app.quit", "<Primary>q");

    /* Default settings */
    priv->csd = TRUE;
    priv->dark_mode = FALSE;
}

/**
 * boilerplate_application_handle_local_command_line:
 * @application:  The application
 * @options: GOptionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Application Sequence: 02
 * Handles the commandline flags like --debug, --version, -v, etc... that do not
 * have a variable set as a pointer in GOptionEntry option_entries[].
 *
 * So if I had an entry like so:
 * { "example", 'e', 0, G_OPTION_ARG_NONE, &EXAMPLE, "example desc", NULL },
 *
 * It would set the value of the --example/-e flag into the EXAMPLE constant
 *
 * But if I had
 * { "example", 'e', 0, G_OPTION_ARG_NONE, NULL, "example desc", NULL },
 *
 * It will be passed into the handle_local_command_line, where it can be access
 * here with g_variant_dict_lookup:
 *
 * g_variant_dict_lookup (options, "example", "b", &priv->example);
 *
 * Returns: -1 Continue / 0 proccess <something>
 *
 * Since: 0.1.0
 */
static gint
boilerplate_application_handle_local_command_line (GApplication *application,
                                                   GVariantDict *options)
{
    g_debug ("Application handling local command line");

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);

    g_variant_dict_lookup (options, "disable-csd", "b", &priv->csd);
    g_variant_dict_lookup (options, "dark-mode", "b", &priv->dark_mode);

    if (g_variant_dict_contains (options, "debug"))
    {
        //g_log_set_handler (NULL, G_LOG_LEVEL_DEBUG, log_handler, NULL);
        g_log (NULL, G_LOG_LEVEL_INFO, "Debugging enabled");
    }

    if (g_variant_dict_contains (options, "version"))
    {
        g_print ("%s\n", PACKAGE_VERSION);
        return 0;
    }

    return -1;
}

/**
 * boilerplate_application_init:
 * @self: The application
 *
 * Initializes the application and connects the local-options/command-line
 * options to boilerplate_application_handle_local_command_line.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_init (BoilerplateApplication *self)
{
    BoilerplateApplicationPrivate *priv = boilerplate_application_get_instance_private (self);
    GSettingsSchemaSource *settings_schema_source;
    GSettingsSchema *settings_schema;

    g_debug ("Application init");

    /* Set application name */
	g_set_application_name (PACKAGE_NAME);

    g_debug ("Application name: %s", g_get_application_name ());

    /* Set default icon via name */
	gtk_window_set_default_icon_name (PACKAGE_ID);

    g_debug ("Application icon name: %s", gtk_window_get_default_icon_name ());

    /* Add option entries / command line flags */
	g_application_add_main_option_entries (G_APPLICATION (self), option_entries);

    /* Connect settings */
    g_debug ("Loading application settings");

    //TODO: Provide a fallback or some kind of cmd switch --settings ./
    settings_schema_source = g_settings_schema_source_get_default ();
    settings_schema = g_settings_schema_source_lookup (settings_schema_source,
                                                       PACKAGE_ID,
                                                       FALSE);

    if (settings_schema == NULL)
    {
        g_info ("Unable to find gschema with name \"%s\", abort!", PACKAGE_ID);
        g_critical ("Was the gschema file installed or is the directory readable?");
        g_application_quit (G_APPLICATION (self));
        return;
    }

    priv->settings = g_settings_new (PACKAGE_ID);
}

/**
 * boilerplate_application_class_init:
 * @class: The application class
 *
 * Initializes the various functions in the application.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_class_init (BoilerplateApplicationClass *klass)
{
    GApplicationClass *application_class = G_APPLICATION_CLASS (klass);

	application_class->startup = boilerplate_application_startup;
	application_class->activate = boilerplate_application_activate;
    application_class->shutdown = boilerplate_application_shutdown;
    application_class->handle_local_options = boilerplate_application_handle_local_command_line;
}

/**
 * boilerplate_application_new:
 *
 * Application Sequence: 01
 * Creates a new GtkApplication.
 *
 * Returns: BoilerplateApplication -> GtkApplication
 *
 * Since: 0.1.0
 */
BoilerplateApplication *
boilerplate_application_new (void)
{
    return g_object_new (BOILERPLATE_TYPE_APPLICATION,
                         "application-id", PACKAGE_ID,
                         "flags", G_APPLICATION_FLAGS_NONE,
                         "resource-base-path", PACKAGE_RESOURCES,
                         NULL);
}

