#define G_LOG_DOMAIN "boilerplate-window"

#include "config.h"

#include <gtk/gtk.h>

#include "boilerplate-application.h"
#include "boilerplate-window.h"

typedef struct
{
    GSettings *settings;

    gboolean csd;
    gboolean dark_theme;

    GtkWidget *toolbar;
    GtkWidget *headerbar;
    GtkWidget *content;
} BoilerplateWindowPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateWindow, boilerplate_window, GTK_TYPE_APPLICATION_WINDOW)

enum {
    PROP_0,
    PROP_CSD,
    PROP_DARK_THEME
};

/**
 * boilerplate_window_toggle_csd:
 * @self: The window
 * @state: TRUE use Client Side Decorations (CSD)
 *         FALSE use traditional title & menubar
 *
 * Toggles the classes CSD state. TRUE uses GNOME/GTK3 headerbars, FALSE uses a
 * more traditional titlebar with ribbon menu.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_toggle_csd (BoilerplateWindow *self,
                               gboolean           state)
{
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    /*
     * If we want the CSD we set the priv->csd state to TRUE but flip the value
     * for gtk_application_window_set_show_menubar (), since we don't want to
     * show the menubar with CSD. Hence the !state
     */
    priv->csd = state;

    gtk_application_window_set_show_menubar (GTK_APPLICATION_WINDOW (self), !state);

    /*
     * Now enable/disable the headerbar on the window.
     */
    boilerplate_window_toggle_csd_appearance (self, state);
}

/**
 * boilerplate_window_toggle_csd_appearance:
 * @self: The window
 * @state: TRUE use Client Side Decorations (CSD)
 *         FALSE use traditional title & menubar
 *
 * Toggles the windows appearance of CSD
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_toggle_csd_appearance (BoilerplateWindow *self,
                                          gboolean           state)
{
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);
    GtkWidget *toolbar_parent;

    toolbar_parent = gtk_widget_get_parent (GTK_WIDGET (priv->toolbar));

    /*
     * Remove toolbar from current parent so we can move it.
     */
    if (toolbar_parent != NULL) {
        gtk_container_remove (GTK_CONTAINER (toolbar_parent), priv->toolbar);
    }

    /*
     * Add toolbar to new container
     */
    if (state) {
        gtk_container_add (GTK_CONTAINER (priv->headerbar), priv->toolbar);
    } else {
        gtk_container_add (GTK_CONTAINER (priv->content), priv->toolbar);

        /*
         * Simply removing the toolbar from the headerbar won't actully
         * disable headerbars. You must fist set a class of
         * .default-decoration on GtkHeaderBar. This only styles the
         * GTKHeaderBar, which will make it smaller. However to get the
         * proper hight of GtkHeaderBar you then must disable headerbar
         * subtitle's or it continuies to take space.
         */
        GtkStyleContext *context;
        context = gtk_widget_get_style_context (priv->headerbar);
        gtk_style_context_add_class (context, "default-decoration");

        gtk_header_bar_set_has_subtitle (GTK_HEADER_BAR (priv->headerbar),
                                         FALSE);
    }
}

/**
 * boilerplate_window_finalize:
 * @object: The window
 *
 * Does any last minute clean-ups when the class ends
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_finalize (GObject *object)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    g_object_unref (priv->toolbar);

    G_OBJECT_CLASS (boilerplate_window_parent_class)->finalize(object);
}


/**
 * boilerplate_window_set_property:
 * @object: The window
 * @property_id: The id of installed propery, list of properties are found at
 *               the top of the file in the enum {} or in the *_class_init ().
 * @value: The properties' value, either the default or what is passed
 * @pspec:  The specification of the property, used to describe what it does in
 *          the api
 *
 * Adds the properties created in *_class_init () to a virtual class
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_set_property (GObject      *object,
                                 guint         property_id,
                                 const GValue *value,
                                 GParamSpec   *pspec)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    switch (property_id) {
        case PROP_CSD:
            boilerplate_window_toggle_csd (self, g_value_get_boolean (value));
            break;

        case PROP_DARK_THEME:
            priv->dark_theme = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

/**
 * boilerplate_window_get_property:
 * @object: The window
 * @property_id: The id of installed propery, list of properties are found at
 *               the top of the file in the enum {} or in the *_class_init ().
 * @value: The properties' value, either the default or what is passed
 * @pspec:  The specification of the property, used to describe what it does in
 *          the api
 *
 * Gets the properties stored in virutal class
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_get_property (GObject    *object,
                                 guint       property_id,
                                 GValue     *value,
                                 GParamSpec *pspec)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    switch (property_id) {
        case PROP_CSD:
            g_value_set_boolean (value, priv->csd);
            break;

        case PROP_DARK_THEME:
            g_value_set_boolean (value, priv->dark_theme);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

/**
 * boilerplate_window_init:
 * @self: The window
 *
 * Initializes the application window.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_init (BoilerplateWindow *self)
{
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);
    GtkBuilder *builder;

    /* Set default CSD state */
    priv->csd = TRUE;

    /* Center window on screen */
    //TODO: Fix & Load last location from gsettings
    gtk_window_set_position (GTK_WINDOW (self), GTK_WIN_POS_CENTER);

    /* Init window.ui template */
    gtk_widget_init_template (GTK_WIDGET (self));

    /* Set window title */
    gtk_window_set_title (GTK_WINDOW (self), PACKAGE_NAME);

    /* Adding headerbar/titlebar/toolbar to private struct */
    if (priv->toolbar == NULL) {
        builder = gtk_builder_new_from_resource (PACKAGE_RESOURCES "/ui/toolbar.ui");
        priv->toolbar = GTK_WIDGET (gtk_builder_get_object (builder, "toolbar"));

        /* Ref toolbar so we can move it between the headerbar and toolbar */
        g_object_ref (priv->toolbar);

        g_object_unref (builder);
    }

    /* Set default window decorations (a.k.a GtkHeaderBar) */
    gtk_window_set_titlebar (GTK_WINDOW (self), priv->headerbar);
    gtk_container_add (GTK_CONTAINER (priv->headerbar), priv->toolbar);
}

/**
 * boilerplate_window_class_init:
 * @class: The window class
 *
 * Initializes the various functions in the window and links the gtk_widget
 * template loaded from gresource.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_class_init (BoilerplateWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = boilerplate_window_finalize;
    object_class->set_property = boilerplate_window_set_property;
    object_class->get_property = boilerplate_window_get_property;

    /* define properties */
    g_object_class_install_property (
        object_class, PROP_CSD,
        g_param_spec_boolean (
            "csd", "Set CSD state",
            "TRUE/FALSE (Enable/Disable) Client Side Decorations (CSD) for application",
            FALSE,
            G_PARAM_READWRITE
        )
    );

    //TODO: Link dark-theme code to property
    g_object_class_install_property (
        object_class, PROP_DARK_THEME,
        g_param_spec_boolean (
            "dark-theme", "Dark Theme",
            "Enable dark theme for application",
            FALSE,
            G_PARAM_READWRITE
        )
    );

    gtk_widget_class_set_template_from_resource (GTK_WIDGET_CLASS (klass),
                                                 PACKAGE_RESOURCES "/ui/window.ui");

    gtk_widget_class_bind_template_child_private (GTK_WIDGET_CLASS (klass), BoilerplateWindow, headerbar);
    gtk_widget_class_bind_template_child_private (GTK_WIDGET_CLASS (klass), BoilerplateWindow, content);
}

/**
 * boilerplate_window_new:
 * @application: The application
 *
 * Creates a new GtkApplicationWindow.
 *
 * Returns: BoilerplateWindow -> GtkApplicationWindow
 *
 * Since: 0.1.0
 */
BoilerplateWindow *
boilerplate_window_new (BoilerplateApplication *application)
{
    return g_object_new (BOILERPLATE_TYPE_WINDOW,
                         "application", application,
                         "title", PACKAGE_NAME,
                         NULL);
}


/******************************************************************************
 *                                   API                                      *
 ******************************************************************************/
void
boilerplate_window_set_csd (BoilerplateWindow *widget,
                            gboolean           value)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    g_return_if_fail (BOILERPLATE_IS_WINDOW (widget));

    boilerplate_window_toggle_csd (self, value);
}

gboolean
boilerplate_window_get_csd (BoilerplateWindow *widget)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    //TODO: Why does g_return_val_if_fail throw error here?
    //g_return_val_if_fail (BOILERPLATE_IS_WINDOW (widget), NULL);

    return (gboolean) priv->csd;
}
